@@ -8,7 +8,7 @@
  *
  *    In some countries OX, OX Open-Xchange, open xchange and OXtender
  *    as well as the corresponding Logos OX Open-Xchange and OX are registered
- *    trademarks of the Open-Xchange, Inc. group of companies.
+ *    trademarks of the OX Software GmbH group of companies.
  *    The use of the Logos is not covered by the GNU General Public License.
  *    Instead, you are allowed to use these Logos according to the terms and
  *    conditions of the Creative Commons License, Version 2.5, Attribution,
@@ -28,7 +28,7 @@
  *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
  *    given Attribution for the derivative code and a license granting use.
  *
- *     Copyright (C) 2004-2013 Open-Xchange, Inc.
+ *     Copyright (C) 2016-2020 OX Software GmbH
  *     Mail: info@open-xchange.com
  *
  *
@@ -73,11 +73,12 @@ import com.hazelcast.core.IMap;
 import com.hazelcast.query.Predicate;
 import com.hazelcast.query.SqlPredicate;
 import com.openexchange.exception.OXException;
+import com.openexchange.java.BoolReference;
 import com.openexchange.server.ServiceExceptionCode;
 import com.openexchange.session.Session;
 import com.openexchange.sessionstorage.SessionStorageExceptionCodes;
 import com.openexchange.sessionstorage.SessionStorageService;
-import com.openexchange.sessionstorage.hazelcast.portable.PortableSession;
+import com.openexchange.sessionstorage.hazelcast.serialization.PortableSession;
 import com.openexchange.threadpool.ThreadPools;
 
 /**
@@ -101,6 +102,34 @@ public class HazelcastSessionStorageService implements SessionStorageService {
         REFERENCE.set(hazelcast);
     }
 
+    private static interface RemovedSessionHandler {
+
+        void handleRemovedSession(PortableSession removedSession);
+    }
+
+    private static final RemovedSessionHandler NOOP_REMOVED_SESSION_HANDLER = new RemovedSessionHandler() {
+
+        @Override
+        public void handleRemovedSession(PortableSession removedSession) {
+            // Nothing
+        }
+    };
+
+    private static final class CollectingRemovedSessionHandler implements RemovedSessionHandler {
+
+        final List<Session> removedSessions;
+
+        CollectingRemovedSessionHandler(int capacity) {
+            super();
+            removedSessions = new ArrayList<Session>(capacity);
+        }
+
+        @Override
+        public void handleRemovedSession(PortableSession removedSession) {
+            removedSessions.add(removedSession);
+        }
+    }
+
     // ----------------------------------------------------------------------------------------------------------------------------------------------- //
 
     private final String sessionsMapName;
@@ -333,72 +362,41 @@ public class HazelcastSessionStorageService implements SessionStorageService {
         }
     }
 
+    /**
+     * {@inheritDoc}
+     */
     @Override
-    public Session[] removeUserSessions(final int userId, final int contextId) throws OXException {
+    public List<Session> removeSessions(List<String> sessionIds) throws OXException {
+        int size;
+        if (null == sessionIds || 0 == (size = sessionIds.size())) {
+            return Collections.emptyList();
+        }
+        /*
+         * check if still active
+         */
         ensureActive();
         /*
-         * search sessions by context- and user-ID
+         * remove sessions...
          */
         IMap<String, PortableSession> sessions = sessions();
-        Set<String> sessionIDs =
-            sessions.keySet(new SqlPredicate(PortableSession.PARAMETER_CONTEXT_ID + " = " + contextId + " AND " + PortableSession.PARAMETER_USER_ID + " = " + userId));
-        if (null == sessionIDs || 0 == sessionIDs.size()) {
-            return new Session[0];
-        }
         /*
          * schedule remove operations
          */
-        Map<String, Future<PortableSession>> futures = new HashMap<String, Future<PortableSession>>(sessionIDs.size());
-        for (String sessionID : sessionIDs) {
-            futures.put(sessionID, sessions.removeAsync(sessionID));
-        }
-        /*
-         * collect removed sessions
-         */
-        List<Session> removedSessions = new ArrayList<Session>(sessionIDs.size());
-        for (Entry<String, Future<PortableSession>> future : futures.entrySet()) {
-            try {
-                PortableSession removedSession = future.getValue().get();
-                if (null != removedSession) {
-                    removedSessions.add(removedSession);
-                } else {
-                    LOG.debug("Session with ID '{}' not found, unable to remove from storage.", future.getKey());
-                }
-            } catch (InterruptedException e) {
-                Thread.currentThread().interrupt();
-                throw SessionStorageExceptionCodes.REMOVE_FAILED.create(e, future.getKey());
-            } catch (ExecutionException e) {
-                final Throwable cause = e.getCause();
-                if (cause instanceof HazelcastInstanceNotActiveException) {
-                    throw handleNotActiveException((HazelcastInstanceNotActiveException) cause);
-                }
-                if (cause instanceof HazelcastException) {
-                    throw handleHazelcastException((HazelcastException) cause, SessionStorageExceptionCodes.REMOVE_FAILED.create(ThreadPools.launderThrowable(e, HazelcastException.class), future.getKey()));
-                }
-
-                // Launder...
-                throw SessionStorageExceptionCodes.REMOVE_FAILED.create(ThreadPools.launderThrowable(e, HazelcastException.class), future.getKey());
-            }
-        }
-        return removedSessions.toArray(new Session[removedSessions.size()]);
+        return removeSessionsByIds(sessionIds, size, sessions);
     }
 
-    @Override
-    public void removeContextSessions(final int contextId) throws OXException {
-        ensureActive();
-        /*
-         * search sessions by context ID
-         */
-        IMap<String, PortableSession> sessions = sessions();
-        Set<String> sessionIDs = sessions.keySet(new SqlPredicate(PortableSession.PARAMETER_CONTEXT_ID + " = " + contextId));
-        if (null == sessionIDs || 0 == sessionIDs.size()) {
-            return;
-        }
+    private List<Session> removeSessionsByIds(Collection<String> sessionIds, int size, IMap<String, PortableSession> sessions) throws OXException {
+        CollectingRemovedSessionHandler removedSessionHandler = new CollectingRemovedSessionHandler(size);
+        removeSessionsByIds(sessionIds, size, removedSessionHandler, sessions);
+        return removedSessionHandler.removedSessions;
+    }
+
+    private void removeSessionsByIds(Collection<String> sessionIds, int size, RemovedSessionHandler removedSessionHandler, IMap<String, PortableSession> sessions) throws OXException {
         /*
          * schedule remove operations
          */
-        Map<String, Future<PortableSession>> futures = new HashMap<String, Future<PortableSession>>(sessionIDs.size());
-        for (String sessionID : sessionIDs) {
+        Map<String, Future<PortableSession>> futures = new HashMap<String, Future<PortableSession>>(size);
+        for (String sessionID : sessionIds) {
             futures.put(sessionID, sessions.removeAsync(sessionID));
         }
         /*
@@ -409,6 +407,8 @@ public class HazelcastSessionStorageService implements SessionStorageService {
                 PortableSession removedSession = future.getValue().get();
                 if (null == removedSession) {
                     LOG.debug("Session with ID '{}' not found, unable to remove from storage.", future.getKey());
+                } else {
+                    removedSessionHandler.handleRemovedSession(removedSession);
                 }
             } catch (InterruptedException e) {
                 Thread.currentThread().interrupt();
@@ -428,6 +428,83 @@ public class HazelcastSessionStorageService implements SessionStorageService {
         }
     }
 
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public Session[] removeLocalUserSessions(int userId, int contextId) throws OXException {
+        ensureActive();
+        /*
+         * search sessions by context- and user-ID
+         */
+        IMap<String, PortableSession> sessions = sessions();
+        Set<String> sessionIDs = sessions.localKeySet(new SqlPredicate(PortableSession.PARAMETER_CONTEXT_ID + " = " + contextId + " AND " + PortableSession.PARAMETER_USER_ID + " = " + userId));
+        int size;
+        if (null == sessionIDs || 0 == (size = sessionIDs.size())) {
+            return new Session[0];
+        }
+        List<Session> removedSessions = removeSessionsByIds(sessionIDs, size, sessions);
+        return removedSessions.toArray(new Session[removedSessions.size()]);
+    }
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public Session[] removeUserSessions(final int userId, final int contextId) throws OXException {
+        ensureActive();
+        /*
+         * search sessions by context- and user-ID
+         */
+        IMap<String, PortableSession> sessions = sessions();
+        Set<String> sessionIDs = sessions.keySet(new SqlPredicate(PortableSession.PARAMETER_CONTEXT_ID + " = " + contextId + " AND " + PortableSession.PARAMETER_USER_ID + " = " + userId));
+        int size;
+        if (null == sessionIDs || 0 == (size = sessionIDs.size())) {
+            return new Session[0];
+        }
+        List<Session> removedSessions = removeSessionsByIds(sessionIDs, size, sessions);
+        return removedSessions.toArray(new Session[removedSessions.size()]);
+    }
+
+    /**
+     * {@inheritDoc}
+     */
+    @Override
+    public void removeLocalContextSessions(int contextId) throws OXException {
+        ensureActive();
+        /*
+         * search sessions by context ID
+         */
+        IMap<String, PortableSession> sessions = sessions();
+        Set<String> sessionIDs = sessions.localKeySet(new SqlPredicate(PortableSession.PARAMETER_CONTEXT_ID + " = " + contextId));
+        int size;
+        if (null == sessionIDs || 0 == (size = sessionIDs.size())) {
+            return;
+        }
+        /*
+         * schedule remove operations
+         */
+        removeSessionsByIds(sessionIDs, size, NOOP_REMOVED_SESSION_HANDLER, sessions);
+    }
+
+    @Override
+    public void removeContextSessions(final int contextId) throws OXException {
+        ensureActive();
+        /*
+         * search sessions by context ID
+         */
+        IMap<String, PortableSession> sessions = sessions();
+        Set<String> sessionIDs = sessions.keySet(new SqlPredicate(PortableSession.PARAMETER_CONTEXT_ID + " = " + contextId));
+        int size;
+        if (null == sessionIDs || 0 == (size = sessionIDs.size())) {
+            return;
+        }
+        /*
+         * schedule remove operations
+         */
+        removeSessionsByIds(sessionIDs, size, NOOP_REMOVED_SESSION_HANDLER, sessions);
+    }
+
     @Override
     public boolean hasForContext(final int contextId) throws OXException {
         ensureActive();
